@@ -333,7 +333,7 @@ static int cmd_device_ready(int argc, char **argv)
 
 	clean_args_no_options(argc, argv, cmd_device_ready_usage);
 
-	if (check_argc_min(argc - optind, 1))
+	if (check_argc_exact(argc - optind, 1))
 		usage(cmd_device_ready_usage);
 
 	fd = open("/dev/btrfs-control", O_RDWR);
@@ -447,6 +447,17 @@ static int cmd_device_stats(int argc, char **argv)
 
 			canonical_path = canonicalize_path((char *)path);
 
+			/* No path when device is missing. */
+			if (!canonical_path) {
+				canonical_path = malloc(32);
+				if (!canonical_path) {
+					error("not enough memory for path buffer");
+					goto out;
+				}
+				snprintf(canonical_path, 32,
+					 "devid:%llu", args.devid);
+			}
+
 			if (args.nr_items >= BTRFS_DEV_STAT_WRITE_ERRS + 1)
 				printf("[%s].write_io_errs   %llu\n",
 				       canonical_path,
